import base64
from typing import List

import cloudpickle
from cryptography.fernet import Fernet

import prefect
from prefect.environments import Environment


class LocalEnvironment(Environment):
    """
    LocalEnvironment is an encrypted and serializable environment for simply packaging
    up flows so they can be stored and transported.

    Args:
        - encryption_key (bytes): key to use in serialization or deserialization of the environment
        - serialized_flow (bytes): a Prefect Flow object that is serialized
    """

    def __init__(self, encryption_key: bytes = None, serialized_flow: bytes = None):
        if encryption_key is None:
            encryption_key = Fernet.generate_key()
        else:
            try:
                Fernet(encryption_key)
            except Exception:
                raise ValueError("Invalid encryption key.")

        self.encryption_key = encryption_key
        self.serialized_flow = serialized_flow

    def build(self, flow: "prefect.Flow") -> "LocalEnvironment":
        """
        Build the LocalEnvironment. Returns a LocalEnvironment with a serialized flow attribute.

        Args:
            - flow (Flow): The prefect Flow object to build the environment for

        Returns:
            - LocalEnvironment: a LocalEnvironment with a serialized flow attribute
        """
        return LocalEnvironment(
            encryption_key=self.encryption_key,
            serialized_flow=self.serialize_flow_to_bytes(flow),
        )

    def serialize_flow_to_bytes(self, flow: "prefect.Flow") -> bytes:
        """
        Serializes a Flow to binary.

        Args:
            - flow (Flow): the Flow to serialize

        Returns:
            - bytes: the serialized Flow
        """
        pickled_flow = cloudpickle.dumps(flow)
        encrypted_pickle = Fernet(self.encryption_key).encrypt(pickled_flow)
        encoded_pickle = base64.b64encode(encrypted_pickle)
        return encoded_pickle

    def deserialize_flow_from_bytes(self, serialized_flow: bytes) -> "prefect.Flow":
        """
        Deserializes a Flow to binary.

        Args:
            - serialized_flow (bytes): the Flow to deserialize

        Returns:
            - Flow: the deserialized Flow
        """
        decoded_pickle = base64.b64decode(serialized_flow)
        decrypted_pickle = Fernet(self.encryption_key).decrypt(decoded_pickle)
        flow = cloudpickle.loads(decrypted_pickle)
        return flow

    def run(self, runner_kwargs: dict = None) -> "prefect.engine.state.State":
        """
        Runs the `Flow` represented by this environment.

        Args:
            - runner_kwargs (dict): Any arguments for `FlowRunner.run()`.

        Returns:
            - State: the state from the flow run
        """

        runner_kwargs = runner_kwargs or {}

        if not self.serialized_flow:
            raise ValueError(
                "No serialized flow found! Has this environment been built?"
            )
        flow = self.deserialize_flow_from_bytes(self.serialized_flow)

        runner_cls = prefect.engine.get_default_flow_runner_class()
        runner = runner_cls(flow=flow)
        return runner.run(**runner_kwargs)



import json
from typing import Any, List

import requests

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


class GetRepoInfo(Task):
    """
    Task for retrieving GitHub repository information using the v3 version of the GitHub REST API.

    Args:
        - repo (str, optional): the name of the repository to open the issue in; must be provided in the
            form `organization/repo_name` or `user/repo_name`; can also be provided to the `run` method
        - info_keys (List[str], optional): a list of repo attributes to pull (e.g., `["stargazers_count", "subscribers_count"]`).
            A full list of available keys can be found in the official [GitHub documentation](https://developer.github.com/v3/repos/)
        - token_secret (str, optional): the name of the Prefect Secret containing your GitHub Access Token;
            defaults to "GITHUB_ACCESS_TOKEN"
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task init method
    """

    def __init__(
        self,
        repo: str = None,
        info_keys: List[str] = None,
        token_secret: str = "GITHUB_ACCESS_TOKEN",
        **kwargs: Any
    ):
        self.repo = repo
        self.info_keys = info_keys or []
        self.token_secret = token_secret
        super().__init__(**kwargs)

    @defaults_from_attrs("repo", "info_keys")
    def run(self, repo: str = None, info_keys: List[str] = None) -> None:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a Flow context,
        or by using `Task.bind`.

        Args:
            - repo (str, optional): the name of the repository to open the issue in; must be provided in the
                form `organization/repo_name`; defaults to the one provided at initialization
            - info_keys (List[str], optional): a list of repo attributes to pull (e.g., `["stargazers_count", "subscribers_count"]`).
                A full list of available keys can be found in the official [GitHub documentation](https://developer.github.com/v3/repos/)

        Raises:
            - ValueError: if a `repo` was never provided
            - HTTPError: if the GET request returns a non-200 status code

        Returns:
            - dict: dictionary of the requested information
        """
        if repo is None:
            raise ValueError("A GitHub repository must be provided.")

        ## prepare the request
        token = Secret(self.token_secret).get()
        url = "https://api.github.com/repos/{}".format(repo)
        headers = {
            "AUTHORIZATION": "token {}".format(token),
            "Accept": "application/vnd.github.v3+json",
        }

        ## send the request
        resp = requests.get(url, headers=headers)
        resp.raise_for_status()
        data = resp.json()

        return {key: data[key] for key in info_keys}



"""
Result Handlers provide the hooks that Prefect uses to store task results in production; a `ResultHandler` can be provided to a `Flow` at creation.

Anytime a task needs its output or inputs stored, a result handler is used to determine where this data should be stored (and how it can be retrieved).
"""
import base64
import tempfile
from typing import Any, Optional

import cloudpickle

from prefect import config
from prefect.client.client import Client
from prefect.engine.result_handlers import ResultHandler


class CloudResultHandler(ResultHandler):
    """
    Hook for storing and retrieving task results from Prefect cloud storage.

    Args:
        - result_handler_service (str, optional): the location of the service
            which will further process and store the results; if not provided, will default to
            the value of `cloud.result_handler` in your config file
    """

    def __init__(self, result_handler_service: str = None) -> None:
        self._client = None  # type: Optional[Client]
        if result_handler_service is None:
            self.result_handler_service = config.cloud.result_handler
        else:
            self.result_handler_service = result_handler_service
        super().__init__()

    def _initialize_client(self) -> None:
        """
        Helper method for ensuring that CloudHandlers which are initialized locally
        do not attempt to start a Client.  This is important because CloudHandlers are
        currently attached to `Flow` objects which need to be serialized / deserialized
        independently of cloud settings.

        This will instantiate a Client upon the first call to (de)serialize.
        """
        if self._client is None:
            self._client = Client()

    def read(self, uri: str) -> Any:
        """
        Read a result from the given URI location.

        Args:
            - uri (str): the path to the location of a result

        Returns:
            - the deserialized result from the provided URI
        """
        self._initialize_client()

        self.logger.debug("Starting to read result from {}...".format(uri))
        res = self._client.get(  # type: ignore
            "/", server=self.result_handler_service, **{"uri": uri}
        )

        try:
            return_val = cloudpickle.loads(base64.b64decode(res.get("result", "")))
        except EOFError:
            return_val = None
        self.logger.debug("Finished reading result from {}...".format(uri))

        return return_val

    def write(self, result: Any) -> str:
        """
        Write the provided result to Prefect Cloud.

        Args:
            - result (Any): the result to store

        Returns:
            - str: the URI path to the result in Cloud storage
        """
        self._initialize_client()

        binary_data = base64.b64encode(cloudpickle.dumps(result)).decode()
        self.logger.debug(
            "Starting to upload result to {}...".format(self.result_handler_service)
        )
        res = self._client.post(  # type: ignore
            "/", server=self.result_handler_service, **{"result": binary_data}
        )
        self.logger.debug(
            "Finished uploading result to {}...".format(self.result_handler_service)
        )
        return res.get("uri", "")



import json
from typing import Any, Dict

from marshmallow import ValidationError, fields, post_load

from prefect.engine.cloud.result_handler import CloudResultHandler
from prefect.engine.result_handlers import (
    GCSResultHandler,
    JSONResultHandler,
    LocalResultHandler,
    ResultHandler,
    S3ResultHandler,
)
from prefect.utilities.serialization import (
    JSONCompatible,
    ObjectSchema,
    OneOfSchema,
    to_qualified_name,
)


class BaseResultHandlerSchema(ObjectSchema):
    class Meta:
        object_class = ResultHandler


class CloudResultHandlerSchema(BaseResultHandlerSchema):
    class Meta:
        object_class = CloudResultHandler

    result_handler_service = fields.String(allow_none=True)


class GCSResultHandlerSchema(BaseResultHandlerSchema):
    class Meta:
        object_class = GCSResultHandler

    bucket = fields.String(allow_none=False)


class JSONResultHandlerSchema(BaseResultHandlerSchema):
    class Meta:
        object_class = JSONResultHandler


class LocalResultHandlerSchema(BaseResultHandlerSchema):
    class Meta:
        object_class = LocalResultHandler

    dir = fields.String(allow_none=True)


class S3ResultHandlerSchema(BaseResultHandlerSchema):
    class Meta:
        object_class = S3ResultHandler

    bucket = fields.String(allow_none=False)


class ResultHandlerSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "ResultHandler": BaseResultHandlerSchema,
        "GCSResultHandler": GCSResultHandlerSchema,
        "S3ResultHandler": S3ResultHandlerSchema,
        "CloudResultHandler": CloudResultHandlerSchema,
        "JSONResultHandler": JSONResultHandlerSchema,
        "LocalResultHandler": LocalResultHandlerSchema,
    }



import warnings
from typing import Any, Callable, List

import prefect
from prefect import Task
from prefect.engine.result import NoResultType
from prefect.triggers import all_finished


class FilterTask(Task):
    """
    Task for filtering lists of results.  The default filter is to filter out `NoResult`s and
    Exceptions for filtering out mapped results.  Note that this task has a default trigger of
    `all_finished` and `skip_on_upstream_skip=False`.

    Args:
        - filter_func (Callable, optional): a function to use for filtering
            results; this function should accept a single positional argument and return a boolean
            indicating whether this result should be _kept_ or not.  The default is
            to filter out `NoResult`s and Exceptions
        - **kwargs (optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(self, filter_func: Callable = None, **kwargs) -> None:
        kwargs.setdefault("skip_on_upstream_skip", False)
        kwargs.setdefault("trigger", all_finished)
        self.filter_func = filter_func or (
            lambda r: not isinstance(r, (NoResultType, Exception))
        )
        super().__init__(**kwargs)

    def run(self, task_results: List[Any]) -> List[Any]:
        """
        Task run method.

        Args:
            - task_results (List[Any]): a list of results from upstream tasks,
                which will be filtered using `self.filter_func`

        Returns:
            - List[Any]: a filtered list of results
        """
        return [r for r in task_results if self.filter_func(r)]



from typing import TYPE_CHECKING, Any

import marshmallow
from marshmallow import fields, post_load

import prefect
from prefect.utilities.collections import DotDict
from prefect.utilities.serialization import (
    Bytes,
    JSONCompatible,
    ObjectSchema,
    OneOfSchema,
    to_qualified_name,
)

if TYPE_CHECKING:
    import prefect.environments.kubernetes


class LocalEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = prefect.environments.LocalEnvironment

    encryption_key = Bytes(allow_none=True)
    serialized_flow = Bytes(allow_none=True)


class DockerEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = prefect.environments.DockerEnvironment

    base_image = fields.String(required=True)
    registry_url = fields.String(required=True)
    image_name = fields.String(allow_none=True)
    image_tag = fields.String(allow_none=True)


# Kubernetes Environments


class DockerOnKubernetesEnvironmentSchema(DockerEnvironmentSchema):
    class Meta:
        object_class = prefect.environments.kubernetes.DockerOnKubernetesEnvironment


class DaskOnKubernetesEnvironmentSchema(DockerEnvironmentSchema):
    class Meta:
        object_class = prefect.environments.kubernetes.DaskOnKubernetesEnvironment

    max_workers = fields.Integer(allow_none=True)


class EnvironmentSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "DockerEnvironment": DockerEnvironmentSchema,
        "LocalEnvironment": LocalEnvironmentSchema,
        "DockerOnKubernetesEnvironment": DockerOnKubernetesEnvironmentSchema,
        "DaskOnKubernetesEnvironment": DaskOnKubernetesEnvironmentSchema,
    }



import atexit
import logging
import os
import queue
import time
from logging.handlers import QueueHandler, QueueListener
from typing import Any

import prefect
from prefect.configuration import config


class RemoteHandler(logging.StreamHandler):
    def __init__(self) -> None:
        super().__init__()
        self.logger_server = config.cloud.log
        self.client = None
        self.errored_out = False

    def emit(self, record) -> None:  # type: ignore
        try:
            if self.errored_out is True:
                return
            if self.client is None:
                from prefect.client import Client

                self.client = Client()  # type: ignore
            assert isinstance(self.client, Client)  # mypy assert
            r = self.client.post(path="", server=self.logger_server, **record.__dict__)
        except:
            self.errored_out = True


old_factory = logging.getLogRecordFactory()


def cloud_record_factory(*args: Any, **kwargs: Any) -> Any:
    record = old_factory(*args, **kwargs)
    record.flow_run_id = prefect.context.get("flow_run_id", "")  # type: ignore
    record.task_run_id = prefect.context.get("task_run_id", "")  # type: ignore
    return record


def configure_logging(testing: bool = False) -> logging.Logger:
    """
    Creates a "prefect" root logger with a `StreamHandler` that has level and formatting
    set from `prefect.config`.

    Args:
        - testing (bool, optional): a boolean specifying whether this configuration
            is for testing purposes only; this helps us isolate any global state during testing
            by configuring a "prefect-test-logger" instead of the standard "prefect" logger

    Returns:
        - logging.Logger: a configured logging object
    """
    name = "prefect-test-logger" if testing else "prefect"
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(config.logging.format)
    formatter.converter = time.gmtime  # type: ignore
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(config.logging.level)

    # send logs to server
    if config.logging.log_to_cloud:
        logging.setLogRecordFactory(cloud_record_factory)
        remote_handler = RemoteHandler()
        logger.addHandler(remote_handler)

    return logger


prefect_logger = configure_logging()


def get_logger(name: str = None) -> logging.Logger:
    """
    Returns a "prefect" logger.

    Args:
        - name (str): if `None`, the root Prefect logger is returned. If provided, a child
            logger of the name `"prefect.{name}"` is returned. The child logger inherits
            the root logger's settings.

    Returns:
        - logging.Logger: a configured logging object with the appropriate name
    """
    if name is None:
        return prefect_logger
    else:
        return prefect_logger.getChild(name)



from prefect.environments.environment import Environment
from prefect.environments.local import LocalEnvironment
from prefect.environments.docker import DockerEnvironment

import prefect.environments.kubernetes



import json

import prefect
from prefect.environments import Environment


def from_file(path: str) -> "Environment":
    """
    Loads a serialized Environment class from a file

    Args:
        - path (str): the path of a file to deserialize as an Environment

    Returns:
        - Environment: an Environment object
    """
    schema = prefect.serialization.environment.EnvironmentSchema()
    with open(path, "r") as f:
        return schema.load(json.load(f))



from typing import TYPE_CHECKING, Any

import marshmallow
from marshmallow import fields

import prefect
from prefect.utilities.serialization import (
    DateTimeTZ,
    ObjectSchema,
    OneOfSchema,
    to_qualified_name,
)

if TYPE_CHECKING:
    import prefect.schedules


class IntervalScheduleSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.IntervalSchedule

    start_date = DateTimeTZ(required=True)
    end_date = DateTimeTZ(allow_none=True)
    interval = fields.TimeDelta(precision="microseconds", required=True)


class CronScheduleSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.CronSchedule

    start_date = DateTimeTZ(allow_none=True)
    end_date = DateTimeTZ(allow_none=True)
    cron = fields.String(required=True)


class OneTimeScheduleSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.OneTimeSchedule

    start_date = DateTimeTZ(required=True)


class ScheduleSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "IntervalSchedule": IntervalScheduleSchema,
        "CronSchedule": CronScheduleSchema,
        "OneTimeSchedule": OneTimeScheduleSchema,
    }



import logging
import uuid
from os import path
from typing import List

import docker
import yaml

import prefect
from prefect.environments import DockerEnvironment


class DockerOnKubernetesEnvironment(DockerEnvironment):
    """
    DockerOnKubernetes is an environment which deploys your image of choice on Kubernetes.
    *Note*: Make sure the base image is able to pip install Prefect. The default image for this
    environment is Python 3.6.

    (A future environment will allow for a minimal set up which does not require pip)

    There are no set up requirements, and execute creates a single job that has the role
    of running the flow. The job created in the execute function does have the requirement in that
    it needs to have an `identifier_label` set with a UUID so resources can be cleaned up
    independently of other deployments.

    Args:
        - base_image (string, optional): the base image for this environment (e.g. `python:3.6`), defaults to `python:3.6`
        - registry_url (string, optional): URL of a registry to push the image to; image will not be pushed if not provided
        - python_dependencies (List[str], optional): list of pip installable dependencies for the image
        - image_name (string, optional): name of the image to use when building, defaults to a UUID
        - image_tag (string, optional): tag of the image to use when building, defaults to a UUID
        - env_vars (dict, optional): a dictionary of environment variables to use when building
        - files (dict, optional): a dictionary of files to copy into the image when building
    """

    def __init__(
        self,
        base_image: str = "python:3.6",
        registry_url: str = None,
        python_dependencies: List[str] = None,
        image_name: str = None,
        image_tag: str = None,
        env_vars: dict = None,
        files: dict = None,
    ) -> None:
        super().__init__(
            base_image=base_image,
            registry_url=registry_url,
            python_dependencies=python_dependencies,
            image_name=image_name,
            image_tag=image_tag,
            env_vars=env_vars,
            files=files,
        )

        self.identifier_label = str(uuid.uuid4())

    def _populate_yaml(self, yaml_obj: dict) -> dict:
        """
        Populate a yaml object used in this environment with the proper values

        Args:
            - yaml_obj (dict): A dictionary representing the parsed yaml

        Returns:
            - dict: a dictionary with the yaml values replaced
        """
        # set identifier labels
        yaml_obj["metadata"]["name"] = "prefect-local-job-{}".format(
            self.identifier_label
        )
        yaml_obj["metadata"]["labels"]["identifier"] = self.identifier_label
        yaml_obj["spec"]["template"]["metadata"]["labels"][
            "identifier"
        ] = self.identifier_label

        # set environment variables
        env = yaml_obj["spec"]["template"]["spec"]["containers"][0]["env"]

        env[0]["value"] = prefect.config.cloud.graphql
        env[1]["value"] = prefect.config.cloud.log
        env[2]["value"] = prefect.config.cloud.result_handler
        env[3]["value"] = prefect.config.cloud.auth_token
        env[4]["value"] = prefect.context.get("flow_run_id", "")

        # set image
        yaml_obj["spec"]["template"]["spec"]["containers"][0]["image"] = "{}:{}".format(
            path.join(self.registry_url, self.image_name),  # type: ignore
            self.image_tag,
        )

        return yaml_obj

    def execute(self) -> None:
        """
        Create a single Kubernetes job on the default namespace that runs a flow
        """
        from kubernetes import client, config

        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            raise EnvironmentError("Environment not currently inside a cluster")

        batch_client = client.BatchV1Api()

        with open(path.join(path.dirname(__file__), "job.yaml")) as job_file:
            job = yaml.safe_load(job_file)
            job = self._populate_yaml(yaml_obj=job)

            # Create Job
            batch_client.create_namespaced_job(namespace="default", body=job)

    def setup(self) -> None:
        """
        No setup is required for this environment
        """
        pass

    def build(
        self, flow: "prefect.Flow", push: bool = True
    ) -> "prefect.environments.DockerEnvironment":
        """
        Build the Docker container. Returns a DockerEnvironment with the appropriate
        image_name and image_tag set.

        Args:
            - flow (prefect.Flow): Flow to be placed in container
            - push (bool): Whether or not to push to registry after build

        Returns:
            - DockerEnvironment: a DockerEnvironment that represents the provided flow.
        """

        image_name, image_tag = self.build_image(flow=flow, push=push)

        return DockerOnKubernetesEnvironment(
            base_image=self.base_image,
            registry_url=self.registry_url,
            image_name=image_name,
            image_tag=image_tag,
        )



from prefect.environments.kubernetes.docker.docker_on_kubernetes import (
    DockerOnKubernetesEnvironment,
)

from prefect.environments.kubernetes.dask.dask_on_kubernetes import (
    DaskOnKubernetesEnvironment,
)



import uuid

from google.cloud import storage
from google.cloud.exceptions import NotFound
from google.oauth2.service_account import Credentials

from prefect import context
from prefect.client import Secret
from prefect.core import Task
from prefect.utilities.tasks import defaults_from_attrs


class GCSBaseTask(Task):
    def __init__(
        self,
        bucket,
        blob=None,
        project=None,
        credentials_secret=None,
        create_bucket=False,
        encryption_key_secret=None,
        **kwargs
    ):
        self.bucket = bucket
        self.blob = blob
        self.project = project
        self.create_bucket = create_bucket
        if credentials_secret is None:
            self.credentials_secret = "GOOGLE_APPLICATION_CREDENTIALS"
        else:
            self.credentials_secret = credentials_secret
        self.encryption_key_secret = encryption_key_secret
        super().__init__(**kwargs)

    def _load_client(self, project, credentials_secret):
        "Creates and returns a GCS Client instance"
        creds = Secret(credentials_secret).get()
        credentials = Credentials.from_service_account_info(creds)
        project = project or credentials.project_id
        client = storage.Client(project=project, credentials=credentials)
        return client

    def _retrieve_bucket(self, client, bucket, create_bucket):
        "Retrieves a bucket based on user settings"
        try:
            bucket = client.get_bucket(bucket)
        except NotFound as exc:
            if create_bucket is True:
                bucket = client.create_bucket(bucket)
            else:
                raise exc
        return bucket

    def _get_blob(self, bucket, blob, encryption_key_secret):
        "Retrieves blob based on user settings."
        if blob is None:
            blob = "prefect-" + context.get("task_run_id", "no-id-" + str(uuid.uuid4()))

        ## pull encryption_key if requested
        if encryption_key_secret is not None:
            encryption_key = Secret(encryption_key_secret).get()
        else:
            encryption_key = None

        return bucket.blob(blob, encryption_key=encryption_key)


class GCSDownload(GCSBaseTask):
    """
    Task template for downloading data from Google Cloud Storage as a string.

    Args:
        - bucket (str): default bucket name to download from; can be overwritten at runtime
        - blob (str, optional): default blob name to download; can be
            overwritten at runtime.  Required prior to running the task.
        - project (str, optional): default Google Cloud project to work within; can be overwritten at runtime.
            If not provided, will be inferred from your Google Cloud credentials
        - credentials_secret (str, optional): the name of the Prefect Secret
            which stores a JSON representation of your Google Cloud credentials; can be overwritten at runtime.
            Defaults to `GOOGLE_APPLICATION_CREDENTIALS`.
        - encryption_key_secret (str, optional): the name of the Prefect Secret
            storing an optional `encryption_key` to be used when downloading the Blob; can be overwritten at runtime
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor

    Note that the design of this task allows you to initialize a _template_ with default settings.  Each inidividual
    occurence of the task in a Flow can overwrite any of these default settings for custom use (for example, if you want to pull different
    credentials for a given Task, or specify the Blob name at runtime).
    """

    def __init__(
        self,
        bucket,
        blob=None,
        project=None,
        credentials_secret=None,
        encryption_key_secret=None,
        **kwargs
    ):
        super().__init__(
            bucket=bucket,
            blob=blob,
            project=project,
            credentials_secret=credentials_secret,
            encryption_key_secret=encryption_key_secret,
            **kwargs
        )

    @defaults_from_attrs(
        "blob", "bucket", "project", "credentials_secret", "encryption_key_secret"
    )
    def run(
        self,
        blob=None,
        bucket=None,
        project=None,
        credentials_secret=None,
        encryption_key_secret=None,
    ):
        """
        Run method for this Task.  Invoked by _calling_ this Task after initialization within a Flow context.

        Args:
            - blob (str, optional): blob name to download from; if not provided here or at initialization,
                a `ValueError` will be raised
            - bucket (str, optional): the bucket name to upload to; if not provided,
                will default to the one provided at initialization
            - project (str, optional): Google Cloud project to work within.
                If not provided here or at initialization, will be inferred from your Google Cloud credentials
            - credentials_secret (str, optional): the name of the Prefect Secret
                which stores a JSON represenation of your Google Cloud credentials; if not provided here or at initialization,
                `GOOGLE_APPLICATION_CREDENTIALS` will be used.
            - encryption_key_secret (str, optional): the name of the Prefect Secret
                storing an optional `encryption_key` to be used when uploading the Blob; if not provided here, will default to the
                one provided at initilization

        Raises:
            - google.cloud.exception.NotFound: if `create_bucket=False` and the bucket name is not found
            - ValueError: if `blob` name hasn't been provided

        Returns:
            - str: the data from the blob, as a string
        """
        ## create client
        client = self._load_client(project, credentials_secret)

        ## retrieve bucket
        bucket = self._retrieve_bucket(
            client=client, bucket=bucket, create_bucket=False
        )

        ## identify blob name
        gcs_blob = self._get_blob(bucket, blob, encryption_key_secret)
        data = gcs_blob.download_as_string()
        return data


class GCSUpload(GCSBaseTask):
    """
    Task template for uploading data to Google Cloud Storage.  Requires the data already be a string.

    Args:
        - bucket (str): default bucket name to upload to; can be overwritten at runtime
        - blob (str, optional): default blob name to upload to; can be overwritten at runtime,
            else a random string beginning with `prefect-` and containing the Task Run ID will be used
        - project (str, optional): default Google Cloud project to work within; can be overwritten at runtime.
            If not provided, will be inferred from your Google Cloud credentials
        - credentials_secret (str, optional): the name of the Prefect Secret
            which stores a JSON represenation of your Google Cloud credentials; can be overwritten at runtime.
            Defaults to `GOOGLE_APPLICATION_CREDENTIALS`.
        - create_bucket (bool, optional): boolean specifying whether to create the bucket if it does not exist,
            otherwise an Exception is raised; can be overwritten at runtime. Defaults to `False`.
        - encryption_key_secret (str, optional): the name of the Prefect Secret
            storing an optional `encryption_key` to be used when uploading the Blob; can be overwritten at runtime
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor

    Note that the design of this task allows you to initialize a _template_ with default settings.  Each inidividual
    occurence of the task in a Flow can overwrite any of these default settings for custom use (for example, if you want to pull different
    credentials for a given Task, or specify the Blob name at runtime).
    """

    def __init__(
        self,
        bucket,
        blob=None,
        project=None,
        credentials_secret=None,
        create_bucket=False,
        encryption_key_secret=None,
        **kwargs
    ):
        super().__init__(
            bucket=bucket,
            blob=blob,
            project=project,
            credentials_secret=credentials_secret,
            create_bucket=create_bucket,
            encryption_key_secret=encryption_key_secret,
            **kwargs
        )

    @defaults_from_attrs(
        "bucket",
        "blob",
        "project",
        "create_bucket",
        "credentials_secret",
        "encryption_key_secret",
    )
    def run(
        self,
        data,
        bucket=None,
        blob=None,
        project=None,
        credentials_secret=None,
        create_bucket=False,
        encryption_key_secret=None,
    ):
        """
        Run method for this Task.  Invoked by _calling_ this Task after initialization within a Flow context.

        Args:
            - data (str): the data to upload; must already be represented as a string
            - bucket (str, optional): the bucket name to upload to; if not provided,
                will default to the one provided at initialization
            - blob (str, optional): blob name to upload to; if not provided here or at initialization,
                a string beginning with `prefect-` and containing the Task Run ID will be used
            - project (str, optional): Google Cloud project to work within.
                If not provided here or at initialization, will be inferred from your Google Cloud credentials
            - credentials_secret (str, optional): the name of the Prefect Secret
                which stores a JSON represenation of your Google Cloud credentials; if not provided here or at initialization,
                `GOOGLE_APPLICATION_CREDENTIALS` will be used.
            - create_bucket (bool, optional): boolean specifying whether to create the bucket if it does not exist,
                otherwise an Exception is raised; if not provided here, will default to the value provided at initialization.
                Defaults to `False`.
            - encryption_key_secret (str, optional): the name of the Prefect Secret
                storing an optional `encryption_key` to be used when uploading the Blob; if not provided here, will default to the
                one provided at initilization

        Raises:
            - google.cloud.exception.NotFound: if `create_bucket=False` and the bucket name is not found

        Returns:
            - str: the blob name that now stores the provided data
        """
        ## create client
        client = self._load_client(project, credentials_secret)

        ## retrieve bucket
        bucket = self._retrieve_bucket(
            client=client, bucket=bucket, create_bucket=create_bucket
        )

        ## identify blob name
        gcs_blob = self._get_blob(bucket, blob, encryption_key_secret)
        gcs_blob.upload_from_string(data)
        return gcs_blob.name



"""
Environments are JSON-serializable objects that fully describe how to run a flow. Serialization
schemas are contained in `prefect.serialization.environment.py`.

Different Environment objects correspond to different computation environments -- for
example, a `LocalEnvironment` runs a flow in the local process; a `DockerEnvironment`
runs a flow in a Docker container.

Environments that are written on top of a type of infrastructure also define how to
set up and execute that environment. e.g. DockerOnKubernetes is an environment which
runs a flow on Kubernetes using the DockerEnvironment.

Some of the information that the environment requires to run a flow -- such as the flow
itself -- may not available when the Environment class is instantiated. Therefore, Environments
may be created with a subset of their (ultimate) required information, and the rest can be
provided when the environment's `build()` method is called.

The most basic Environment is a `LocalEnvironment`. This class stores a serialized version
of a Flow and deserializes it to run it. It is expected that most other Environments
will manipulate LocalEnvironments to actually run their flows. For example, the
`DockerEnvironment` builds a Docker image with all necessary dependencies installed
and also a serialized `LocalEnvironment`. When the `DockerEnvironment` is deployed,
the container in turn runs the `LocalEnvironment`.
"""

import json
from typing import TYPE_CHECKING

import prefect

if TYPE_CHECKING:
    import prefect.engine.state


class Environment:
    """
    Base class for Environments.

    An environment is an object that can be instantiated in a way that makes it possible to
    call `environment.run()` and run a flow.

    Because certain `__init__` parameters may not be known when the Environment is first
    created, including which Flow to run, Environments have a `build()` method that takes
    a `Flow` argument and returns an Environment with all `__init__` parameters specified.

    The setup and execute functions are limited to environments which have an infrastructure
    requirement. However the build and run functions are limited to base environments such
    as (LocalEnvironment and DockerEnvironment) from which infrastructure dependent environments
    inherit from.
    """

    def __init__(self) -> None:
        pass

    def build(self, flow: "prefect.Flow") -> "Environment":
        """
        Builds the environment for a specific flow. A new environment is returned.

        Args:
            - flow (prefect.Flow): the Flow for which the environment will be built

        Returns:
            - Environment: a new environment that can run the provided Flow.
        """
        raise NotImplementedError()

    def execute(self) -> None:
        """
        Executes the environment on any infrastructure created during setup
        """
        pass

    def run(self) -> "prefect.engine.state.State":
        """
        Runs the `Flow` represented by this environment.

        Returns:
            - prefect.engine.state.State: the state of the flow run
        """
        raise NotImplementedError()

    def setup(self) -> None:
        """
        Sets up the infrastructure needed for this environment
        """
        pass

    def serialize(self) -> dict:
        """
        Returns a serialized version of the Environment

        Returns:
            - dict: the serialized Environment
        """
        schema = prefect.serialization.environment.EnvironmentSchema()
        return schema.dump(self)

    def to_file(self, path: str) -> None:
        """
        Serialize the environment to a file.

        Args:
            - path (str): the file path to which the environment will be written
        """
        with open(path, "w") as f:
            json.dump(self.serialize(), f)



"""
Triggers are functions that determine if task state should change based on
the state of preceding tasks.
"""
from typing import Callable, Set

from prefect import context
from prefect.engine import signals, state


def all_finished(upstream_states: Set["state.State"]) -> bool:
    """
    This task will run no matter what the upstream states are, as long as they are finished.

    Args:
        - upstream_states (set[State]): the set of all upstream states
    """
    if not all(s.is_finished() for s in upstream_states):
        raise signals.TRIGGERFAIL(
            'Trigger was "all_finished" but some of the upstream tasks were not finished.'
        )

    return True


def manual_only(upstream_states: Set["state.State"]) -> bool:
    """
    This task will never run automatically, because this trigger will
    always place the task in a Paused state. The only exception is if
    the "resume" keyword is found in the Prefect context, which happens
    automatically when a task starts in a Resume state.

    Args:
        - upstream_states (set[State]): the set of all upstream states
    """
    if context.get("resume"):
        return True

    raise signals.PAUSE('Trigger function is "manual_only"')


def all_successful(upstream_states: Set["state.State"]) -> bool:
    """
    Runs if all upstream tasks were successful. Note that `SKIPPED` tasks are considered
    successes and `TRIGGER_FAILED` tasks are considered failures.

    Args:
        - upstream_states (set[State]): the set of all upstream states
    """

    if not all(s.is_successful() for s in upstream_states):
        raise signals.TRIGGERFAIL(
            'Trigger was "all_successful" but some of the upstream tasks failed.'
        )
    return True


def all_failed(upstream_states: Set["state.State"]) -> bool:
    """
    Runs if all upstream tasks failed. Note that `SKIPPED` tasks are considered successes
    and `TRIGGER_FAILED` tasks are considered failures.

    Args:
        - upstream_states (set[State]): the set of all upstream states
    """

    if not all(s.is_failed() for s in upstream_states):
        raise signals.TRIGGERFAIL(
            'Trigger was "all_failed" but some of the upstream tasks succeeded.'
        )
    return True


def any_successful(upstream_states: Set["state.State"]) -> bool:
    """
    Runs if any tasks were successful. Note that `SKIPPED` tasks are considered successes
    and `TRIGGER_FAILED` tasks are considered failures.

    Args:
        - upstream_states (set[State]): the set of all upstream states
    """

    if not any(s.is_successful() for s in upstream_states):
        raise signals.TRIGGERFAIL(
            'Trigger was "any_successful" but none of the upstream tasks succeeded.'
        )
    return True


def any_failed(upstream_states: Set["state.State"]) -> bool:
    """
    Runs if any tasks failed. Note that `SKIPPED` tasks are considered successes and
    `TRIGGER_FAILED` tasks are considered failures.

    Args:
        - upstream_states (set[State]): the set of all upstream states
    """

    if not any(s.is_failed() for s in upstream_states):
        raise signals.TRIGGERFAIL(
            'Trigger was "any_failed" but none of the upstream tasks failed.'
        )
    return True


# aliases
always_run = all_finished  # type: Callable[[Set["state.State"]], bool]



import datetime
import multiprocessing
import signal
import threading
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

import dask
import dask.bag

import prefect
from prefect.core.edge import Edge

if TYPE_CHECKING:
    import prefect.engine.runner
    import prefect.engine.state
    from prefect.engine.state import State
StateList = Union["State", List["State"]]


class Heartbeat(threading.Timer):
    def run(self) -> None:
        self.finished.wait(self.interval)  # type: ignore
        while not self.finished.is_set():  # type: ignore
            self.function(*self.args, **self.kwargs)  # type: ignore
            self.finished.wait(self.interval)  # type: ignore


def run_with_heartbeat(
    runner_method: Callable[..., "prefect.engine.state.State"]
) -> Callable[..., "prefect.engine.state.State"]:
    """
    Utility decorator for running class methods with a heartbeat.  The class should implement
    `self._heartbeat` with no arguments.
    """

    @wraps(runner_method)
    def inner(
        self: "prefect.engine.runner.Runner", *args: Any, **kwargs: Any
    ) -> "prefect.engine.state.State":
        try:
            timer = Heartbeat(prefect.config.cloud.heartbeat_interval, self._heartbeat)
            timer.daemon = True
            try:
                self._heartbeat()
            except:
                pass
            timer.start()
            return runner_method(self, *args, **kwargs)
        finally:
            timer.cancel()

    return inner


def multiprocessing_timeout(
    fn: Callable, *args: Any, timeout: int = None, **kwargs: Any
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by spawning a new multiprocess.Process() and joining with timeout.

    Args:
        - fn (callable): the function to execute
        - *args (Any): arguments to pass to the function
        - timeout (int): the length of time to allow for
            execution before raising a `TimeoutError`, represented as an integer in seconds
        - **kwargs (Any): keyword arguments to pass to the function

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - AssertionError: if run from a daemonic process
        - TimeoutError: if function execution exceeds the allowed timeout
    """

    if timeout is None:
        return fn(*args, **kwargs)

    def retrieve_value(
        *args: Any, _container: multiprocessing.Queue, **kwargs: Any
    ) -> None:
        """Puts the return value in a multiprocessing-safe container"""
        try:
            _container.put(fn(*args, **kwargs))
        except Exception as exc:
            _container.put(exc)

    q = multiprocessing.Queue()  # type: multiprocessing.Queue
    kwargs["_container"] = q
    p = multiprocessing.Process(target=retrieve_value, args=args, kwargs=kwargs)
    p.start()
    p.join(timeout)
    p.terminate()
    if not q.empty():
        res = q.get()
        if isinstance(res, Exception):
            raise res
        return res
    else:
        raise TimeoutError("Execution timed out.")


def main_thread_timeout(
    fn: Callable, *args: Any, timeout: int = None, **kwargs: Any
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by setting a `signal` alarm on a timer. Must be run in the main thread.

    Args:
        - fn (callable): the function to execute
        - *args (Any): arguments to pass to the function
        - timeout (int): the length of time to allow for
            execution before raising a `TimeoutError`, represented as an integer in seconds
        - **kwargs (Any): keyword arguments to pass to the function

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - TimeoutError: if function execution exceeds the allowed timeout
        - ValueError: if run from outside the main thread
    """

    if timeout is None:
        return fn(*args, **kwargs)

    def error_handler(signum, frame):  # type: ignore
        raise TimeoutError("Execution timed out.")

    try:
        signal.signal(signal.SIGALRM, error_handler)
        signal.alarm(timeout)
        return fn(*args, **kwargs)
    finally:
        signal.alarm(0)



#!/usr/bin/env python


import click
import json
import logging
import os
import requests
import sys
import prefect


@click.group()
def cli():
    """
    The Prefect CLI
    """
    pass


@cli.command()
def make_user_config():
    """
    Generates a user configuration file
    """
    user_config_path = prefect.config.get("user_config_path")
    if not user_config_path:
        raise ValueError("No user config path set!")
    elif os.path.isfile(user_config_path):
        raise ValueError("A file already exists at {}".format(user_config_path))

    os.makedirs(os.path.dirname(user_config_path), exist_ok=True)
    with open(user_config_path, "w") as user_config:
        user_config.write(
            "# This is a user configuration file.\n"
            "# Settings placed here will overwrite Prefect's defaults."
        )

    click.secho("Config created at {}".format(user_config_path), fg="green")


@cli.command()
@click.argument("environment_file", type=click.Path(exists=True))
@click.option("--runner_kwargs", default={})
def run(environment_file, runner_kwargs):
    """
    Run a flow from an environment file.
    """
    schema = prefect.serialization.environment.EnvironmentSchema()
    with open(environment_file, "r") as f:
        environment = schema.load(json.load(f))

    click.echo(environment.run(runner_kwargs=runner_kwargs))


@cli.command()
@click.argument("environment_metadata")
def create_environment(environment_metadata):
    """
    Call the setup and execute functions for a given environment.
    """
    schema = prefect.serialization.environment.EnvironmentSchema()
    environment = schema.load(json.loads(environment_metadata))

    environment.setup()
    environment.execute()



import filecmp
import json
import logging
import os
import shutil
import tempfile
import textwrap
import uuid
from typing import Iterable, List

import docker

import prefect
from prefect.environments import Environment, LocalEnvironment
from prefect.utilities.exceptions import SerializationError


class DockerEnvironment(Environment):
    """
    This is a base environment which takes a flow, serializes it into a LocalEnvironment,
    and places it inside of a Docker image. This image is then used in any environment
    which depends on using Docker containers (e.g. the Kubernetes environments).

    Args:
        - base_image (string): the base image for this environment (e.g. `python:3.6`)
        - registry_url (string, optional): URL of a registry to push the image to; image will not be pushed if not provided
        - python_dependencies (List[str], optional): list of pip installable dependencies for the image
        - image_name (string, optional): name of the image to use when building, defaults to a UUID
        - image_tag (string, optional): tag of the image to use when building, defaults to a UUID
        - env_vars (dict, optional): a dictionary of environment variables to use when building
        - files (dict, optional): a dictionary of files to copy into the image when building
    """

    def __init__(
        self,
        base_image: str,
        registry_url: str = None,
        python_dependencies: List[str] = None,
        image_name: str = None,
        image_tag: str = None,
        env_vars: dict = None,
        files: dict = None,
    ) -> None:
        self.base_image = base_image
        self.registry_url = registry_url
        self.image_name = image_name
        self.image_tag = image_tag
        self.python_dependencies = python_dependencies or []
        self.env_vars = env_vars or {}
        self.files = files or {}
        not_absolute = [
            file_path for file_path in self.files if not os.path.isabs(file_path)
        ]
        if not_absolute:
            raise ValueError(
                "Provided paths {} are not absolute file paths, please provide absolute paths only.".format(
                    ", ".join(not_absolute)
                )
            )

    def _parse_generator_output(self, generator: Iterable) -> None:
        """
        Parses and writes a Docker command's output to stdout
        """
        for item in generator:
            item = item.decode("utf-8")
            for line in item.split("\n"):
                if line:
                    output = json.loads(line).get("stream")
                    if output and output != "\n":
                        print(output.strip("\n"))

    def build(
        self, flow: "prefect.Flow", push: bool = True
    ) -> "prefect.environments.DockerEnvironment":
        """
        Build the Docker image. Returns a DockerEnvironment with the appropriate
        image_name and image_tag set.

        Args:
            - flow (prefect.Flow): Flow to be placed the image
            - push (bool): Whether or not to push to registry after build

        Returns:
            - DockerEnvironment: a DockerEnvironment that represents the provided flow.
        """

        image_name, image_tag = self.build_image(flow=flow, push=push)

        return DockerEnvironment(
            base_image=self.base_image,
            registry_url=self.registry_url,
            image_name=image_name,
            image_tag=image_tag,
        )

    def build_image(self, flow: "prefect.Flow", push: bool = True) -> tuple:
        """
        Build the Docker image using the docker python library. Optionally pushes the
        image if both `push`=`True` and `self.registry_url` is set.

        Args:
            - flow (prefect.Flow): Flow to be placed the image
            - push (bool): Whether or not to push to registry after build

        Returns:
            - tuple: `image_name`, `image_tag` (strings)
        """
        image_name = str(uuid.uuid4())
        image_tag = str(uuid.uuid4())

        with tempfile.TemporaryDirectory() as tempdir:

            self.pull_image()

            self.create_dockerfile(flow=flow, directory=tempdir)

            client = docker.APIClient(base_url="unix://var/run/docker.sock")

            if self.registry_url:
                full_name = os.path.join(self.registry_url, image_name)
            elif push is True:
                raise ValueError(
                    "This environment has no `registry_url`, and cannot be pushed."
                )
            else:
                full_name = image_name

            logging.info("Building the flow's container environment...")
            output = client.build(
                path=tempdir, tag="{}:{}".format(full_name, image_tag), forcerm=True
            )
            self._parse_generator_output(output)
            if len(client.images(name=full_name)) == 0:
                raise SerializationError(
                    "Your flow failed to deserialize in the container; please ensure that all necessary files and dependencies have been included."
                )

            if push:
                self.push_image(full_name, image_tag)

                # Remove the image locally after being pushed
                client.remove_image(
                    image="{}:{}".format(full_name, image_tag), force=True
                )

            return image_name, image_tag

    def pull_image(self) -> None:
        """Pull the image specified so it can be built.

        In order for the docker python library to use a base image it must be pulled
        from either the main docker registry or a separate registry that must be set as
        `registry_url` on this class.
        """
        client = docker.APIClient(base_url="unix://var/run/docker.sock")

        output = client.pull(self.base_image, stream=True, decode=True)
        for line in output:
            if line.get("progress"):
                print(line.get("status"), line.get("progress"), end="\r")
        print("")

    def push_image(self, image_name: str, image_tag: str) -> None:
        """Push this environment to a registry

        Args:
            - image_name (str): Name for the image
            - image_tag (str): Tag for the image
        """
        client = docker.APIClient(base_url="unix://var/run/docker.sock")

        logging.info("Pushing image to the registry...")

        output = client.push(image_name, tag=image_tag, stream=True, decode=True)
        for line in output:
            if line.get("progress"):
                print(line.get("status"), line.get("progress"), end="\r")
        print("")

    def create_dockerfile(self, flow: "prefect.Flow", directory: str = None) -> None:
        """Creates a dockerfile to use as the container.

        In order for the docker python library to build a container it needs a
        Dockerfile that it can use to define the container. This function takes the
        image and python_dependencies then writes them to a file called Dockerfile.

        *Note*: if `files` are added to this container, they will be copied to this directory as well.

        Args:
            - flow (Flow): the flow that the container will run
            - directory (str, optional): A directory where the Dockerfile will be created,
                if no directory is specified is will be created in the current working directory
        """

        directory = directory or "./"

        with open(os.path.join(directory, "Dockerfile"), "w+") as dockerfile:

            # Generate RUN pip install commands for python dependencies
            pip_installs = ""
            if self.python_dependencies:
                for dependency in self.python_dependencies:
                    pip_installs += "RUN pip install {}\n".format(dependency)

            env_vars = ""
            if self.env_vars:
                white_space = " " * 20
                env_vars = "ENV " + " \ \n{}".format(white_space).join(
                    "{k}={v}".format(k=k, v=v) for k, v in self.env_vars.items()
                )

            copy_files = ""
            if self.files:
                for src, dest in self.files.items():
                    fname = os.path.basename(src)
                    full_fname = os.path.join(directory, fname)
                    if (
                        os.path.exists(full_fname)
                        and filecmp.cmp(src, full_fname) is False
                    ):
                        raise ValueError(
                            "File {fname} already exists in {directory}".format(
                                fname=full_fname, directory=directory
                            )
                        )
                    else:
                        shutil.copy2(src, full_fname)
                    copy_files += "COPY {fname} {dest}\n".format(fname=fname, dest=dest)

            # Create a LocalEnvironment to run the flow
            # the local environment will be placed in the container and run when the container
            # runs
            local_environment = LocalEnvironment().build(flow=flow)
            flow_path = os.path.join(directory, "flow_env.prefect")
            local_environment.to_file(flow_path)

            # write a healthcheck script into the image
            healthcheck = textwrap.dedent(
                """\
            print('Beginning health check...')
            from prefect.utilities.environments import from_file

            local_env = from_file('/root/.prefect/flow_env.prefect')
            flow = local_env.deserialize_flow_from_bytes(local_env.serialized_flow)
            print('Healthcheck: OK')
            """
            )

            with open(os.path.join(directory, "healthcheck.py"), "w") as health_file:
                health_file.write(healthcheck)

            file_contents = textwrap.dedent(
                """\
                FROM {base_image}

                RUN pip install pip --upgrade
                RUN pip install wheel
                {pip_installs}

                RUN mkdir /root/.prefect/
                COPY flow_env.prefect /root/.prefect/flow_env.prefect
                COPY healthcheck.py /root/.prefect/healthcheck.py
                {copy_files}

                ENV PREFECT_ENVIRONMENT_FILE="/root/.prefect/flow_env.prefect"
                ENV PREFECT__USER_CONFIG_PATH="/root/.prefect/config.toml"
                {env_vars}

                RUN pip install prefect

                RUN python /root/.prefect/healthcheck.py
                """.format(
                    base_image=self.base_image,
                    pip_installs=pip_installs,
                    copy_files=copy_files,
                    env_vars=env_vars,
                    access_token=os.getenv("PERSONAL_ACCESS_TOKEN"),
                )
            )

            dockerfile.write(file_contents)
